"""Utils for email display."""

from collections.abc import Iterable

from django.contrib.auth import get_user_model
from django.utils.safestring import mark_safe

User = get_user_model()

AT, DOT = (f'<span class="{tag}"></span>' for tag in ("at", "dot"))
EMAIL_JS = """<script>
{{
   const mails = {mails}.map((m) => m.join('@')).join();
   const link = '<a {klass} href="mailto:' + mails + {subject} + '">' + {text} + '</a>';
   document.write(link);
}}
</script>
<noscript>{noscript}</noscript>
"""


def flatten(iterable):
    """Flatten an iterable of "values or iterables"."""
    return (
        [c for b in iterable for c in flatten(b)]
        if isinstance(iterable, Iterable) and not isinstance(iterable, str)
        else [iterable]
    )


def show_emails(
    authenticated: bool,
    *to: [User | str],
    text: str = "",
    subject: str = "",
    klass: str = "",
) -> str:
    """Show an email as a link to authenticated users, and obfuscated for others."""
    mails = [r.email if hasattr(r, "email") else r for r in flatten(to)]
    subject = f"?subject={subject}" if subject else ""
    klass = f'class="{klass}"' if klass else ""
    if authenticated:
        mails = ",".join(mails)
        content = f'<a {klass} href="mailto:{mails}{subject}">{text or mails}</a>'
    else:
        content = EMAIL_JS.format(
            mails=[m.split("@") for m in mails],
            noscript=",".join(mails).replace("@", AT).replace(".", DOT),
            text=f"'{text}'" if text else "mails",
            subject=f"'{subject}'",
            klass=klass,
        )
    return mark_safe(f'<span class="mail">{content}</span>')



"""General QuerySets for NDH."""

from django.db.models import QuerySet


class NameOrderedQuerySet(QuerySet):
    """Order a queryset on the "name" field."""

    def name_ordered(self) -> QuerySet:
        """Order by name."""
        return self.order_by("name")



"""General utils for NDH."""

import os
from pathlib import Path
from typing import TYPE_CHECKING, TypeVar, Union

from django.db import models
from django.db.models.functions import Coalesce

if TYPE_CHECKING:  # pragma: no cover
    from django.db.models.manager import RelatedManager

Numeric = TypeVar("Numeric", int, float)


def full_url(url: str = "", domain: None | str = None, protocol: str = "https") -> str:
    """Prepend protocol and domain name to an url."""
    if domain is None:
        from django.contrib.sites.models import Site

        domain = Site.objects.get_current().domain
    return f"{protocol}://{domain}{url}"


def query_sum(
    queryset: Union[models.QuerySet, "RelatedManager"],
    field: str,
    output_field: models.Field | None = None,
) -> Numeric:
    """Let the DBMS perform a sum on a queryset."""
    return queryset.aggregate(
        s=Coalesce(models.Sum(field), 0, output_field=output_field),
    )["s"]


def get_env(env_file: str = ".env") -> None:
    """Set default environment variables from .env file."""

    def load_env(path: Path):
        with path.open() as f:
            for line in f.readlines():
                if line.startswith("#"):
                    continue
                try:
                    key, val = line.split("=", maxsplit=1)
                    os.environ.setdefault(key.strip(), val.strip())
                except ValueError:
                    pass

    current = Path(env_file)
    parent = Path().parent / env_file
    if current.exists():
        load_env(current)
    elif parent.exists():
        load_env(parent)


def set_env(var: str, val: str, env_file: str = ".env") -> None:
    """Set default environment variables from .env file."""

    def update_env(path: Path):
        with path.open() as f:
            lines = f.readlines()

        updated = False

        with path.open("w") as f:
            for line in lines:
                if not line.startswith("#") and line.split("=")[0].strip() == var:
                    print(f"{var} = {val}", file=f)
                    updated = True
                else:
                    print(line.strip(), file=f)
            if not updated:
                print(f"{var} = {val}", file=f)

    current = Path(env_file)
    parent = Path().parent / env_file
    if current.exists():
        update_env(current)
    elif parent.exists():
        update_env(parent)



"""NDH Widgets."""

from django.forms.widgets import DateInput, TextInput

DATE_FORMAT = "%Y-%m-%d"
TYPE_DATE = {"type": "date"}
TYPE_TIME = {"type": "time"}


class AccessibleDateInput(DateInput):
    """Set date HTML type."""

    def __init__(self):
        """Initialize with date type."""
        super().__init__(attrs=TYPE_DATE, format=DATE_FORMAT)


class DatalistInput(TextInput):
    """Datalist widget."""

    template_name = "ndh/forms/widgets/datalist.html"

    def __init__(self, datalist=None, **attrs):
        """Initialize with datalist."""
        self.datalist = datalist
        super().__init__(**attrs)

    def get_context(self, name, value, attrs):
        """Add datalist to the context."""
        context = super().get_context(name, value, attrs)
        context["datalist"] = (
            self.datalist() if callable(self.datalist) else self.datalist
        )
        return context



"""Django template tags for NDH."""

from django import template
from django.contrib.auth import get_user_model
from django.db.models.query import QuerySet
from django.urls import reverse
from django.utils.safestring import mark_safe

from ndh.mail import show_emails

User = get_user_model()

register = template.Library()


TEL_JS = """<script>
{{
   const num = '{phone}';
   const num_nosp = '{phone_nosp}';
   const link = '<a href="tel:' + num_nosp  + '">' + num + '</a>';
   document.write(link);
}}
</script>
<noscript>{phone}</noscript>
"""


@register.simple_tag(takes_context=True)
def show_email(
    context: dict,
    *to: [User | str],
    text: str = "",
    subject: str = "",
    klass: str = "",
) -> str:
    """Show an email as a link to connected users, and obfuscated for others."""
    return show_emails(
        context["request"].user.is_authenticated,
        *to,
        text=text,
        subject=subject,
        klass=klass,
    )


@register.simple_tag(takes_context=True)
def show_phone(context: dict, phone: str) -> str:
    """Show phone number as a link to connected users, and obfuscated for others."""
    phone_nosp = phone.replace(" ", "").replace("(", "").replace(")", "")
    if context["request"].user.is_authenticated:
        content = f'<a href="tel:{phone_nosp}">{phone}</a>'
    else:
        content = TEL_JS.format(phone=phone, phone_nosp=phone_nosp)
    return mark_safe(f'<span class="phone">{content}</span>')


@register.filter
def admin_url(obj) -> str:
    """Get the admin url of a Model or QuerySet instance."""
    if isinstance(obj, QuerySet):  # type: ignore
        obj = obj[0]._meta
        return reverse(f"admin:{obj.app_label}_{obj.model_name}_changelist")
    if hasattr(obj, "_queryset_class"):
        obj = obj.first()._meta
        return reverse(f"admin:{obj.app_label}_{obj.model_name}_changelist")
    return reverse(
        f"admin:{obj._meta.app_label}_{obj._meta.model_name}_change",
        args=[obj.pk],
    )


@register.simple_tag(takes_context=True)
def navbar_item(context, view_name: str, link: str) -> str:
    """Get a navbar item, activated if its url is in the current request path."""
    url = reverse(view_name)
    active = "active" if url == context.request.path else ""
    return mark_safe(
        f'<li class="nav-item me-auto {active}">'
        f'<a class="nav-link" href="{url}">{link}</a></li>',
    )


@register.filter
def user_smcp(user: User):
    """Get user's capitalized `first_name` + capitalized & small-capsed `last_name`."""
    return mark_safe(
        f"{user.first_name.capitalize()} "
        f'<span class="smcp">{user.last_name.capitalize()}</span>',
    )



"""General view Mixins for NDH."""

from django.contrib.auth.mixins import UserPassesTestMixin
from django.views.generic.base import ContextMixin


class AttrContextMixin(ContextMixin):
    """Mixin to add view attributes to context."""

    attr_context = []

    def get_context_data(self, **kwargs):
        """Update context from attributes."""
        return super().get_context_data(
            **{key: getattr(self, key) for key in self.attr_context},
            **kwargs,
        )


class SuperUserRequiredMixin(UserPassesTestMixin):
    """Mixin that allow access only to superusers."""

    def test_func(self):
        """Check that the user has superuser access."""
        return self.request.user.is_superuser


class NDHFormMixin(AttrContextMixin):
    """Mixin setting a default form template and title."""

    template_name = "ndh/base_form.html"
    title = ""
    continue_edit = False
    attr_context = ["title", "continue_edit"]

    def get_success_url(self):
        """Redirect to current page if continue_edit."""
        if self.continue_edit and "continue_edit" in self.request.GET:
            return self.request.path
        return super().get_success_url()


class NDHDeleteMixin(NDHFormMixin):
    """Mixin setting a default template for delete views."""

    template_name = "ndh/base_delete.html"
